import os
import sys
import json
from femagtools import amela


def read_output():
    with open('tests/data/amela.out', 'r') as f:
        data = f.readline()
    return data.split('\n')[0]


def read_json():
    with open('tests/data/pm_data/pm_data_se38.json', 'r') as f:
        data = json.load(f)
    return data


def test_amela():
    al = amela.Amela(workdir='tests/data',
                     magnet_data=dict(name='pm_data'))
    loss = al()
    r = read_json()
    assert r['name'] == 'pm_data_se38'
    assert len(r['phi']) == 73
    assert read_output() == 'succeed'
    assert round(loss['pm_data_se38']['total_loss'], 3) == 0.881



# -*- coding: utf-8 -*-
"""
    femagtools.airgap
    ~~~~~~~~~~~~~~~~~

    Read airgap dat file


"""
import numpy as np
import logging

logger = logging.getLogger(__name__)


def fft(pos, b, pmod=0):
    """calculate fft spectrum of flux density and return samples,
    values, amplitude and phase of base harmonic

    Arguments:
      pos: (list of floats) sample positions
      b: (list of floats) flux density values
      pmod: number of poles in model (ignored if 0)
    """
    model_angle = pos[-1] - pos[0]
    ntiles = int(round(360/model_angle))

    if pmod:
        negative_periodic = pmod % 2
    else:
        negative_periodic = np.abs(b[0] - b[-1])/np.max(b) > 1

    if negative_periodic:
        bx = np.append(
            np.concatenate(
                [n*b[:-1]
                 for n in [m % 2 or -1
                           for m in range(1, ntiles+1)]]),
            b[0])
    else:
        bx = np.append(
            np.tile(b[:-1], ntiles),
            b[0])

    N = len(bx)
    # compute DFT from induction
    Y = np.fft.fft(bx)

    # find the peak (amplitude of base harmonic)
    i = np.argmax(np.abs(Y[:N//2]))
    a = 2*np.abs(Y[i])/N
    freq = np.fft.fftfreq(N, d=pos[1]-pos[0])
    T0 = np.abs(1/freq[i])
    npoles = 2*int(np.ceil(360/T0))
    logger.info("flux density: %s poles B amp %f ",
                npoles, a)

    alfa0 = np.angle(Y[i])
    return dict(Bamp=a, npoles=npoles,
                phi0=alfa0,
                pos=pos.tolist(),
                B=b.tolist(),
                nue=np.arange(0, 9*npoles).tolist(),
                B_nue=(2*np.abs(Y[:9*npoles])/N).tolist(),
                B_fft=(a*np.cos(2*np.pi*pos/T0+alfa0)).tolist(),
                Bi=bx.tolist(),
                phi=np.linspace(pos[0], 360+pos[0], len(bx)).tolist())


def read(filename, pmod=0):
    """read dat file with columns (phi, Br, Bphi)
    returns samples, values, amplitude and phase of base harmonic

    Args:
      filename: the name of the file to be processed
      pmod: number of poles in model (ignored if 0)
    """
    bag = np.loadtxt(filename).T
    if len(bag) < 3:
        logger.warn("%s has incomplete content", filename)
        return(dict())

    return fft(bag[0], bag[1], pmod)



# -*- coding: utf-8 -*-
"""
    femagtools.machine
    ~~~~~~~~~~~~~~~~~~

    Analytical machine models

"""
import numpy as np
from femagtools.bch import Reader
from .pm import PmRelMachineLdq, PmRelMachinePsidq
from .im import InductionMachine
from .utils import betai1, iqd, invpark, K, T, puconv, dqpar_interpol
import copy
import logging

logger = logging.getLogger(__name__)


def __scale_losses(losses, lfe):
    if losses:
        l = {k: lfe*np.array(losses[k]) for k in (
            'styoke_hyst', 'styoke_eddy',
            'stteeth_hyst', 'stteeth_eddy',
            'rotor_hyst', 'rotor_eddy',
            'magnet')}
        l['speed'] = losses['speed']
        return l
    return {}


def create_from_eecpars(temp, eecpars, lfe=1, wdg=1):
    """create machine according to the eecpars:
    PM, EESM or IM"""
    if 'ldq' in eecpars:  # this is a PM (or EESM)
        if (isinstance(eecpars['ldq'], list) and
            'ex_current' in eecpars['ldq'][0] or
            isinstance(eecpars['ldq'], dict) and
                'ex_current' in eecpars['ldq']):
            pars = copy.deepcopy(eecpars)
            pars['tcu1'] = temp[0]
            pars['tcu2'] = temp[1]
            return femagtools.machine.sm.SynchronousMachine(pars, lfe=lfe, wdg=wdg)

        if isinstance(eecpars['ldq'], list) and len(eecpars['ldq']) > 1:
            x, dqp = dqpar_interpol(
                temp[1], eecpars['ldq'], ipkey='temperature')
        else:
            dqp = eecpars['ldq'][0]
            logger.warning(
                "single temperature DQ parameters: unable to fit temperature %s", temp)

        beta = dqp['beta']
        i1 = np.array(dqp['i1'])/wdg
        psid = wdg*lfe*dqp['psid']
        psiq = wdg*lfe*dqp['psiq']
        try:
            losses = __scale_losses(dqp['losses'], lfe)
            losses['ef'] = eecpars['ldq'][-1]['losses']['ef']
            losses['eh'] = eecpars['ldq'][-1]['losses']['ef']
        except KeyError as e:
            logger.warning(e)
            losses = {}

        return PmRelMachineLdq(
            eecpars['m'], eecpars['p'],
            r1=eecpars['r1']*lfe*wdg**2,
            ls=eecpars['ls1']*wdg**2,
            psid=psid,
            psiq=psiq,
            losses=losses,
            beta=beta,
            i1=i1,
            tcu1=temp[0])

    else:  # must be an induction machine (TODO: check scaling)
        pars = copy.deepcopy(eecpars)
        pars['r1'] = lfe*wdg**2*pars['r1']
        pars['lsigma1'] = lfe*pars['lsigma1']
        pars['lsigma2'] = lfe*pars['lsigma2']
        pars['psiref'] = wdg*lfe*pars['psiref']
        pars['u1ref'] = wdg*lfe*pars['u1ref']
        pars['rotor_mass'] = lfe*pars['rotor_mass']
        pars['r2'] = lfe*pars['r2']
        pars['fec'] = lfe*pars['fec']
        pars['fee'] = lfe*pars['fee']
        pars['im'] = [im/wdg for im in pars['im']]
        pars['psi'] = [psi*wdg*lfe for psi in pars['psi']]
        pars['tcu1'] = temp[0]
        pars['tcu2'] = temp[1]
        return InductionMachine(pars)


def __scale_losses(losses, lfe):
    if losses:
        l = {k: lfe*np.array(losses[k]) for k in (
            'styoke_hyst', 'styoke_eddy',
            'stteeth_hyst', 'stteeth_eddy',
            'rotor_hyst', 'rotor_eddy',
            'magnet')}
        l['speed'] = losses['speed']
        return l
    return {}


def create(bch, r1, ls, lfe=1, wdg=1):
    """create PmRelMachine from BCH

    Arguments:
      bch: BchReader or Erg object
      r1: winding resistance
      ls: winding leakage
      lfe: scale factor length
      wdg: scale factor number of windings
"""
    m = 3
    if isinstance(bch, Reader):
        p = bch.machine['p']
        if bch.type.lower().find('psid-psiq-identification') >= 0:
            id = np.array(bch.psidq['id'])/wdg
            iq = np.array(bch.psidq['iq'])/wdg
            psid = wdg*lfe*np.array(bch.psidq['psid'])
            psiq = wdg*lfe*np.array(bch.psidq['psiq'])
            try:
                losses = __scale_losses(bch.psidq['losses'], lfe)
                losses['ef'] = bch.lossPar.get('ef', [2.0, 2.0])
                losses['eh'] = bch.lossPar.get('eh', [1.0, 1.0])
            except KeyError:
                losses = {}
            return PmRelMachinePsidq(m, p, psid, psiq, r1*lfe*wdg**2,
                                     id, iq, ls*wdg**2, losses=losses)

        if bch.type.lower().find('ld-lq-identification') >= 0:
            beta = bch.ldq['beta']
            i1 = np.array(bch.ldq['i1'])/wdg
            psid = wdg*lfe*np.array(bch.ldq['psid'])
            psiq = wdg*lfe*np.array(bch.ldq['psiq'])
            try:
                losses = __scale_losses(bch.ldq['losses'], lfe)
                losses['ef'] = bch.lossPar.get('ef', [2.0, 2.0])
                losses['eh'] = bch.lossPar.get('eh', [1.0, 1.0])
            except KeyError:
                losses = {}
            return PmRelMachineLdq(m, p, psid=psid, psiq=psiq,
                                   r1=r1*lfe*wdg**2,
                                   i1=i1, beta=beta, ls=ls*wdg**2,
                                   losses=losses)
        raise ValueError("Unsupported BCH type {}".format(bch.type))
    # must be ERG type:
    p = int(round(np.sqrt(2)*bch['M_sim'][-1][-1]/(
        m*bch['Psi_d'][-1][-1] * bch['i1'][-1])))

    return PmRelMachineLdq(m, p, r1=r1*lfe*wdg**2,
                           beta=bch['beta'], i1=np.array(bch['i1'])/wdg,
                           psid=wdg*lfe*np.array(bch['Psi_d'])/np.sqrt(2),
                           psiq=wdg*lfe*np.array(bch['Psi_q'])/np.sqrt(2),
                           ls=ls*wdg**2)



import numpy as np
import scipy.interpolate as ip
import logging
from .utils import betai1
from .pm import PmRelMachineLdq
from . import create_from_eecpars

logger = logging.getLogger("femagtools.effloss")


def _generate_mesh(n, T, nb, Tb, npoints):
    """return speed and torque list for driving/braking speed range

    arguments:
    n: list of speed values in driving mode (1/s)
    T: list of torque values in driving mode (Nm)
    nb: list of speed values in braking mode (1/s)
    Tb: list of torque values in braking mode (Nm)
    npoints: number of values for speed and torque list
    """
    if nb:
        nmax = 0.99*min(max(n), max(nb))
        tmin, tmax = min(Tb), max(T)
        tnum = npoints[1]//2
    else:
        nmax = max(n)
        tmin, tmax = 0, max(T)
        tnum = npoints[1]
    tip = ip.interp1d(n, T)
    if nb and Tb:
        tbip = ip.interp1d(nb, Tb)
    else:
        def tbip(x): return 0

    nxtx = []
    for nx in np.linspace(1, nmax, npoints[0]):
        t0 = tbip(nx)
        t1 = tip(nx)
        try:
            npnts = round((t1-t0) / (tmax-tmin) * npoints[1])
            if npnts > 2:
                for t in np.concatenate(
                        (np.linspace(t0, 0.015*tmin, tnum),
                         np.linspace(0.015*tmax, t1, tnum))):
                    nxtx.append((nx, t))
        except ValueError as ex:
            logger.warning("n, t0, t1 %s tmin, tmax %s, npoints %d",
                           (n, t0, t1), (tmin, tmax), npoints[1])
    return np.array(nxtx).T


def efficiency_losses_map(eecpars, u1, T, temp, n, npoints=(50, 40)):
    """return speed, torque efficiency and losses

    arguments:
    eecpars: (dict) EEC Parameter with
      dicts at different temperatures (or machine object)
    u1: (float) phase voltage (V rms)
    T: (float) starting torque (Nm)
    temp: temperature (°C)
    n: (float) maximum speed (1/s)
    npoints: (list) number of values of speed and torque

    """
    if isinstance(eecpars, dict):
        if isinstance(temp, (list, tuple)):
            xtemp = [temp[0], temp[1]]
        else:
            xtemp = [temp, temp]
        m = create_from_eecpars(xtemp, eecpars)
    else:  # must be an instance of Machine
        m = eecpars
    if isinstance(T, list):
        r = {'T': T, 'n': n}
        rb = {'T': [], 'n': []}
    else: # calculate speed,torque charactersics
        nmax = n
        r = m.characteristics(T, nmax, u1)  # driving mode
        if isinstance(m, PmRelMachineLdq):
            if min(m.betarange) >= -np.pi/2:  # driving mode only
                rb = {}
                rb['n'] = None
                rb['T'] = None
            else:
                rb = m.characteristics(-T, max(r['n']), u1)  # braking mode
        else:
            rb = m.characteristics(-T, max(r['n']), u1)  # braking mode

    ntmesh = _generate_mesh(r['n'], r['T'],
                            rb['n'], rb['T'], npoints)

    if isinstance(m, PmRelMachineLdq):
        iqd = np.array([
            m.iqd_torque_umax(
                nt[1],
                2*np.pi*nt[0]*m.p,
                u1)[:-1]
            for nt in ntmesh.T]).T
        beta, i1 = betai1(iqd[0], iqd[1])
        uqd = [m.uqd(2*np.pi*n*m.p, *i)
               for n, i in zip(ntmesh[0], iqd.T)]
        u1 = np.linalg.norm(uqd, axis=1)/np.sqrt(2.0)
        f1 = ntmesh[0]*m.p
    else:
        f1 = []
        u1max = u1
        r = dict(u1=[], i1=[], plfe1=[], plcu1=[], plcu2=[])
        for nx, tq in ntmesh.T:
            wm = 2*np.pi*nx
            w1 = m.w1(u1max, m.psiref, tq, wm)
            f1.append(w1/2/np.pi)
            u1 = m.u1(w1, m.psi, wm)
            r['u1'].append(np.abs(u1))
            i1 = m.i1(w1, m.psi, wm)
            r['i1'].append(np.abs(i1))
            r['plfe1'].append(m.m*np.abs(u1)**2/m.rfe(w1, m.psi))
            i2 = m.i2(w1, m.psi, wm)
            r['plcu1'].append(m.m*np.abs(i1)**2*m.rstat(w1))
            r['plcu2'].append(m.m*np.abs(i2)**2*m.rrot(w1-m.p*wm))


    if isinstance(m, PmRelMachineLdq):
        plfe1 = m.iqd_plfe1(*iqd, f1)
        plfe2 = m.iqd_plfe2(iqd[0], iqd[1], f1)
        plmag = m.iqd_plmag(iqd[0], iqd[1], f1)
        plcu1 = m.iqd_plcu1(iqd[0], iqd[1], 2*np.pi*f1)
        plcu2 = m.iqd_plcu2(*iqd)
        try:
            tfric = eecpars['kfric_b']*eecpars['rotor_mass']*30e-3/np.pi
        except:
            tfric = 0
    else:
        plfe1 = np.array(r['plfe1'])
        plfe2 = np.zeros(ntmesh.shape[1])
        plmag = np.zeros(ntmesh.shape[1])
        plcu1 = np.array(r['plcu1'])
        plcu2 = np.array(r['plcu2'])
        iqd = np.zeros(ntmesh.shape)
        u1 = np.array(r['u1'])
        i1 = np.array(r['i1'])
        tfric = 0

    plfric = 2*np.pi*ntmesh[0]*tfric
    ntmesh[1] -= tfric
    pmech = np.array(
        [2*np.pi*nt[0]*nt[1]
         for nt in ntmesh.T])
    ploss = plfe1+plfe2+plmag+plcu1+plcu2+plfric

    eta = []
    for pm, pl in zip(pmech, ploss):
        p1 = pm+pl
        if (p1 <= 0 and pm >= 0) or (p1 >= 0 and pm <= 0):
            e = 0
        elif abs(p1) > abs(pm):
            e = pm / p1
        else:
            e = p1 / pm
        eta.append(e)

    return dict(
        iq=iqd[0].tolist(),
        id=iqd[1].tolist(),
        i1=i1.tolist(),
        u1=u1.tolist(),
        n=ntmesh[0].tolist(),
        T=ntmesh[1].tolist(),
        pmech=pmech.tolist(),
        eta=eta,
        plfe1=plfe1.tolist(),
        plfe2=plfe2.tolist(),
        plmag=plmag.tolist(),
        plcu1=plcu1.tolist(),
        plcu2=plcu2.tolist(),
        plfric=plfric.tolist(),
        losses=ploss.tolist())



'''
    femagtools.amela
    ~~~~~~~~~~~~~~~~

    Calculate Magnet Losses with AMELA
'''
import sys
import json
import pathlib
import logging
from pathlib import Path
import femagtools.nc
import numpy as np
from numpy import pi
import subprocess

#  set logging
logger = logging.getLogger(__name__)


class Amela():
    '''Run Amela Calculation
    Parameters
    ----------
    workdir : str
        working directory of femag calculation
        (The directory that the nc file)
    amela_dir: str (optional)
        amela directory 
    magnet_data : dict
        calculation control
        -- name must be provided
        dict(name='test',
        -- the following parameters are optional
             mur=1.05,
             sigma=625000,
             hm=3,
             wm=20,
             lm=30,
             speed=1000, 
             nsegwid=0, 
             nseglen=0)
    '''

    def __init__(self, workdir, magnet_data, amela_dir=None):

        self.magn = magnet_data
        self.workdir = pathlib.Path(workdir)
        if amela_dir is not None: 
            self.amela_dir = pathlib.Path(amela_dir)
        else: 
            self.amela_dir = self.workdir

        self.jsonfile = []
        if sys.platform == 'win32':
            self.cmd = [str(self.amela_dir / 'AMELA.BAT')]
        else:
            self.cmd = [str(self.amela_dir / 'AMELA')]
        # default batch 
        self.cmd.append('--b')
        # append calc options
        if 'speed' in self.magn:
            self.cmd.append(f"--speed {self.magn['speed']}")
        if 'nsegwid' in self.magn:
            self.cmd.append(f"--nsegwid {self.magn['nsegwid']}")
        if 'nseglen' in self.magn:
            self.cmd.append(f"--nseglen {self.magn['nseglen']}")

    def get_magnet_data(self):
        '''Extract magnet data from nc file
        Parameters
        ----------
            None
        Returns
        ----------
        pm_data: list of magnet data
        '''
        nc_name = self.workdir / self.magn['name']
        r = femagtools.nc.read(nc_name)

        mag_spels = r.magnet_super_elements()
        spel_key = [i.key for i in mag_spels]
        spel_area = [float(i.area())*1e6 for i in mag_spels]

        pm_elem_key = []
        elem = []
        pm_node_key = [[] for i in range(len(spel_key))]
        bndkey = [[] for i in range(len(spel_key))]
        bx = [[] for i in range(len(spel_key))]
        by = [[] for i in range(len(spel_key))]
        xcp = [[] for i in range(len(spel_key))]
        ycp = [[] for i in range(len(spel_key))]
        bndx = [[] for i in range(len(spel_key))]
        bndy = [[] for i in range(len(spel_key))]

        # conductivity and permeability of the magnets
        cond = 0
        mur = 0
        # read boundary nodes
        for k, i in enumerate(mag_spels):

            cond = i.conduc
            mur = 1/i.elements[0].reluc[0]
            logger.debug('Magnet: mur=%s, conductivity=%s', mur, cond)

            pm_elem_key.append([j.key for j in i.elements])
            elem.append(len(i.elements))
            for j in i.elements:
                for kk in j.vertices:
                    pm_node_key[k].append(kk.key - 1)

            for bnd in i.nodechains:
                for kk in bnd.nodes:
                    if len(bndkey[k]) > 0:
                        if kk.key != bndkey[k][-1]:
                            bndkey[k].append(kk.key)
                            bndx[k].append(kk.x*1e3)
                            bndy[k].append(kk.y*1e3)
                    else:
                        bndkey[k].append(kk.key)
                        bndx[k].append(kk.x*1e3)
                        bndy[k].append(kk.y*1e3)

                bndkey[k].pop(-1)
                bndx[k].pop(-1)
                bndy[k].pop(-1)

        # default load angle (input beta I vs Up)
        num_cases = r.el_fe_induction_1.shape[3] - 1
        if 'loadcase' in self.magn:
            indx = self.magn['loadcase'] - 1
        else:
            indx = num_cases
        if indx == 3:
            indx = num_cases  # avoid error

        # stationary case, no rotation
        poles = 0
        try:
            poles = r.num_poles
        except:
            pass

        # read mesh and flux density
        for i in range(len(spel_key)):
            bx[i].append(np.zeros((elem[i], len(r.pos_el_fe_induction))))
            by[i].append(np.zeros((elem[i], len(r.pos_el_fe_induction))))
            for index, j in enumerate(mag_spels[i].elements):
                xcp[i].append(float(j.center[0]*1e3))
                ycp[i].append(float(j.center[1]*1e3))
                theta = np.arctan2(float(j.center[1]), float(j.center[0]))
                fd = r.flux_density(j, 0, indx)
                if poles == 0:
                    bx[i][0][index, :] = fd['bx']
                    by[i][0][index, :] = fd['by']
                else:
                    bx[i][0][index, :] = fd['bx']*np.cos(theta) - \
                        fd['by']*np.sin(theta)
                    by[i][0][index, :] = fd['bx']*np.sin(theta) + \
                        fd['by']*np.cos(theta)

        if poles == 0:
            freq = self.magn.get('f', r.speed)
            time_vec = np.linspace(0, 1/freq, len(r.pos_el_fe_induction))
            pos = dict(time=time_vec.tolist(),
                       freq=freq,
                       t=float(1/freq))
            # reset num.poles
            poles = 1
        else:
            rpm = self.magn.get('speed', r.speed)
            ag_sim = r.pos_el_fe_induction[-1] - r.pos_el_fe_induction[0]
            pos = dict(phi=(r.pos_el_fe_induction*180/pi).tolist(),
                       speed=rpm,
                       t=float(60/rpm*ag_sim/360))  # TODO
        # prep dictionary for the loss calculation
        pm_data = []
        for i in range(len(spel_key)):
            pm_data.append(dict(name='pm_data_se' + str(spel_key[i]),
                                hm=self.magn.get('hm', 0),
                                wm=self.magn.get('wm', 0),
                                lm=self.magn.get('lm', r.arm_length*1e3),
                                ls=r.arm_length*1e3,
                                sigma=float(self.magn.get('sigma', cond)),
                                mur=float(self.magn.get('mur', mur)),
                                loadcase=self.magn.get('loadcase', indx),
                                numpoles=poles,
                                nodes=dict(),
                                elements=dict(),
                                bndkeys=bndkey[i],
                                bndx=[float(c) for c in bndx[i]],
                                bndy=[float(c) for c in bndy[i]],
                                area=spel_area[i]))
            pm_data[i].update(pos)

        for k in range(len(pm_node_key)):
            for i, j in enumerate(pm_node_key[k]):
                pm_data[k]['nodes'][str(j + 1)] = dict(
                    key=int(j + 1),
                    x=float(r.nodes[j].x*1e3),
                    y=float(r.nodes[j].y*1e3),
                    Az=float(r.nodes[j].vpot[0])
                )
            for i, j in enumerate(pm_elem_key[k]):
                pm_data[k]['elements'][str(j)] = dict(
                    key=int(j + 1),
                    xcp=xcp[k][i],
                    ycp=ycp[k][i],
                    Bx=bx[k][0][i, :].tolist(),
                    By=by[k][0][i, :].tolist()
                )

        if len(mag_spels) / r.poles_sim > 1:
            return pm_data
        else:
            return [pm_data[0]]

    def export_json(self, pm_data):
        '''Export magnet data to json files
        Parameters
        ----------
        pm_data: list of magnet data
        Returns
        ----------
        None
        ----------
        '''
        pm_dir = self.amela_dir / self.magn['name']
        pm_dir.mkdir(exist_ok=True)
        for i in pm_data:
            filename = (pm_dir / i['name']).with_suffix('.json')
            self.jsonfile.append(str(filename))  # for the future use
            with filename.open('w') as f:
                json.dump(i, f)
            logger.info('Exporting %s ...', i['name'])

    def read_loss(self, pm_data):
        '''Read magnet losses
        Parameters
        ----------
            pm_data : dict
        Returns
        ----------
            losses : dict
        ----------
        '''
        losses = {}
        for i in range(len(pm_data)):
            dirname = self.magn['name'] + '/' + pm_data[i]['name'] + \
                '/EClosses.tab'
            result_name = self.amela_dir / dirname
            with result_name.open() as f:
                data = np.loadtxt(f)
                total_loss = np.mean(data[0:-1, -1]) 
                losses[pm_data[i]['name']] = {"loss_data":data, "total_loss": total_loss}
                logger.info("Magnet losses in superelement %s is %.3f W",
                            pm_data[i]['name'].split('se')[-1], total_loss)
        return losses

    def __call__(self, ialh=False):
        '''Run amela calculation
        Parameters
        ----------
            None
        Returns
        ----------
            losses : dict
        ----------
        '''
        # get magnet data
        r = self.get_magnet_data()
        # export to json
        self.export_json(r)
        # run amela 
        calc_method = 'IALH' if ialh else '3DI'
        cmd = self.cmd + ['--calc', calc_method, self.magn['name'] + '/']
        log_file = self.amela_dir / 'amela.out'
        logger.info("Calculating magnet losses with AMELA ...")
        with log_file.open('w') as output:
            subprocess.run(cmd, stdout=output)
        return self.read_loss(r)
        


"""
  femagtools.sm
  ~~~~~~~~~~~~~
  wound-rotor synchronous machine (EESM) electrical circuit model

  Copyright 2022: Semafor Informatik & Energie AG, Switzerland
"""
import numpy as np
import femagtools.bch
import scipy.optimize as so
import scipy.interpolate as ip
from .utils import skin_resistance
import logging
import warnings

EPS = 1e-13

eecdefaults = dict(
    zeta1=0.2,
    zeta2=0,
    gam=0.7,
    kh=2,
    tcu1=20,
    tcu2=20,
    rotor_mass=0,
    kfric_b=1)

logger = logging.getLogger('sm')
logging.captureWarnings(True)


def _linsampl(exc, excl, a):
    """auxiliary func for linear sampling of nonlinear sequence
    arguments:
    exc: (list) nonlinear sequence of excitation current
    excl: (list) linear sequence of excitation current
    a: (array) matrix to be resampled"""
    z = []
    b, i = a.shape[1:]
    for x in range(b):
        for y in range(i):
            zl = ip.interp1d(exc, a[:, x, y], kind='linear')
            z.append(zl(excl))
    return np.array(z).T.reshape(len(excl), b, i)


def _splinterp(beta, i1, betax, i1x, a):
    """auxiliary function to increase resolution of array a
    using a cubic spline interpolation.
    arguments:
    beta: (list) n original up-i angles in rad
    i1: (list) m original current in A
    betax: (list) nx new up-i angles in rad
    i1x: (list) mx new currents
    a: (nxm array) to be interpolated"""
    f = ip.RectBivariateSpline(
        beta, i1, np.asarray(a),
        kx=3, ky=3).ev
    X, Y = np.meshgrid(betax, i1x)
    return f(X, Y).T


def _islinear(exc):
    return np.abs(np.sum(
        np.asarray(exc)**2 -
        np.linspace(exc[0], exc[-1], len(exc))**2)) < 1e-2


def gradient_respecting_bounds(bounds, fun, eps=1e-8):
    """bounds: list of tuples (lower, upper)"""
    def gradient(x):
        logger.info(x)
        fx = fun(x)
        grad = np.zeros(len(x))
        for k in range(len(x)):
            d = np.zeros(len(x))
            d[k] = eps if x[k] + eps <= bounds[k][1] else -eps
            grad[k] = (fun(x + d) - fx) / d[k]
        return grad
    return gradient


class SynchronousMachine(object):
    def __init__(self, pars, lfe=1, wdg=1):
        for k in eecdefaults.keys():
            setattr(self, k, eecdefaults[k])
        for k in pars:
            if k not in ('ldq', ):
                setattr(self, k, pars[k])

        islinear = True
        iexc = [l['ex_current'] for l in pars['ldq']]
        i1 = [i/wdg for i in pars['ldq'][-1]['i1']]
        i1x = np.linspace(i1[0], i1[-1], 12)
        beta = [b*np.pi/180 for b in pars['ldq'][-1]['beta']]
        betax = np.linspace(beta[0], beta[-1], 20)

        if _islinear(iexc):
            psid = wdg*lfe*np.array([
                _splinterp(beta, i1, betax, i1x, l['psid'])
                for l in pars['ldq']])
            psiq = wdg*lfe*np.array([
                _splinterp(beta, i1, betax, i1x, l['psiq'])
                for l in pars['ldq']])
            exc = iexc
        else:
            islinear = False
            nsamples = 10
            iexcl = np.linspace(iexc[0], iexc[-1], nsamples)
            exc = iexcl
            psid = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                [_splinterp(beta, i1, betax, i1x, l['psid'])
                 for l in pars['ldq']]))
            psiq = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                [_splinterp(beta, i1, betax, i1x, l['psiq'])
                 for l in pars['ldq']]))

        self.psidf = ip.RegularGridInterpolator(
            (exc, betax, i1x), np.sqrt(2)*psid, bounds_error=False)
        self.psiqf = ip.RegularGridInterpolator(
            (exc, betax, i1x), np.sqrt(2)*psiq, bounds_error=False)
        i1max = np.sqrt(2)*(max(i1))
        self.bounds = [(np.cos(min(beta))*i1max, i1max),
                       (-i1max, 0),
                       (iexc[0], iexc[-1])]
        self.fo = 50
        self.plexp = {'styoke_hyst': 1.0,
                      'stteeth_hyst': 1.0,
                      'styoke_eddy': 2.0,
                      'stteeth_eddy': 2.0,
                      'rotor_hyst': 1.0,
                      'rotor_eddy': 2.0}
        keys = self.plexp.keys()
        try:
            if islinear:
                pfe = {k: np.array([l['losses'][k]
                                    for l in pars['ldq']])
                       for k in keys}
            else:
                pfe = {k: _linsampl(iexc, iexcl,
                                    np.array([l['losses'][k]
                                             for l in pars['ldq']]))
                       for k in keys}
            self._losses = {k: ip.RegularGridInterpolator(
                (exc, beta, i1), lfe*np.array(pfe[k])) for k in keys}
            self._set_losspar(pars['ldq'][0]['losses']['speed'],
                              pars['ldq'][0]['losses']['ef'],
                              pars['ldq'][0]['losses']['hf'])
        except KeyError:
            logger.warning("loss map missing")
            self._losses = {k: lambda x: 0 for k in (
                'styoke_hyst', 'stteeth_hyst',
                'styoke_eddy', 'stteeth_eddy',
                'rotor_hyst', 'rotor_eddy')}

    def _set_losspar(self, speed, ef, hf):
        self.fo = speed*self.p
        self.plexp = {'styoke_hyst': hf,
                      'stteeth_hyst': hf,
                      'styoke_eddy': ef,
                      'stteeth_eddy': ef,
                      'rotor_hyst': hf,
                      'rotor_eddy': ef}

    def rstat(self, w):
        """stator resistance"""
        return skin_resistance(self.r1, w, self.tcu1, self.zeta1,
                               self.gam, self.kh)

    def rrot(self, w):
        """rotor resistance"""
        return skin_resistance(self.r2, w, self.tcu2, self.zeta2,
                               0.0, 1)

    def psi(self, iq, id, iex):
        """return psid, psiq of currents iq, id"""
        beta = np.arctan2(id, iq)
        if beta > 0:
            beta -= 2*np.pi
        i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
        try:
            return self.psidf((iex, beta, i1)), self.psiqf((iex, beta, i1))
        except ValueError as ex:
            logger.error(iex, iq, id, beta, i1)
            raise ex

    def torque_iqd(self, iq, id, iex):
        "torque at q-d-current"
        psid, psiq = self.psi(iq, id, iex)
        return self.m*self.p/2*(psid*iq - psiq*id)

    def uqd(self, w1, iq, id, iex):
        """return uq, ud of frequency w1 and d-q current"""
        psid, psiq = self.psi(iq, id, iex)
        r1 = self.rstat(w1)
        return r1*id + w1*psid, r1*iq - w1*psiq

    def plcu1(self, iqde, w1):
        r1 = self.rstat(w1)
        return 3/2*r1*(iqde[0]**2 + iqde[1]**2)

    def plcu2(self, iqde, w1=0):
        r2 = self.rrot(0)
        return r2*iqde[2]**2

    def culoss(self, iqde, w1=0):
        r2 = self.rrot(0)
        return self.plcu1(iqde, w1) + self.plcu2(iqde, 0)

    def plfe1(self, beta, i1, iex, f1):
        return np.sum([
            self._losses[k]((iex, beta, i1))*(f1/self.fo)**self.plexp[k][0]
            for k in ('styoke_eddy', 'styoke_hyst',
                      'stteeth_eddy', 'stteeth_hyst')], axis=0)

    def iqd_plcu1(self, iq, id, f1):
        return self.plcu1((iq, id), 2*np.pi*f1)

    def iqd_plcu2(self, iq, id, iex):
        return self.plcu2((iq, id, iex))

    def iqd_plfe1(self, iq, id, iex, f1):
        beta = np.arctan2(id, iq)
        beta[beta > 0] -= 2*np.pi
        i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
        return self.plfe1(beta, i1, iex, f1)

    def iqd_plfe2(self, iq, id, f1):
        return np.zeros(np.asarray(iq).shape)

    def iqd_plmag(self, iq, id, f1):
        return np.zeros(np.asarray(iq).shape)

    def iqd_torque(self, torque, disp=False, maxiter=500):
        """return currents for torque with minimal losses"""
        if torque > 0:
            startvals = self.bounds[0][1], 0, sum(self.bounds[-1])/2
        else:
            startvals = -self.bounds[0][1], 0, sum(self.bounds[-1])/2

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            res = so.minimize(
                self.culoss, startvals, method='SLSQP',  # trust-constr
                bounds=self.bounds,
                #            jac=gradient_respecting_bounds(self.bounds, self.culoss),
                constraints=[
                    {'type': 'eq',
                     'fun': lambda iqd: self.torque_iqd(*iqd) - torque}],
                options={'disp': disp, 'maxiter': maxiter})
        if res['success']:
            return res.x
        logger.warning("%s: torque=%f %f, io=%s",
                       res['message'], torque, self.torque_iqd(*startvals),
                       startvals)
        raise ValueError(res['message'])

    def iqd_torque_umax(self, torque, w1, u1max, io=0,
                        disp=False, maxiter=500):
        """return currents for torque with minimal losses"""
        if io == 0:
            iqde = self.iqd_torque(torque, disp, maxiter)
            if np.linalg.norm(
                    self.uqd(w1, *iqde)) <= u1max*np.sqrt(2):
                return (*iqde, torque)
            io = iqde[0], iqde[1], self.bounds[-1][0]
        res = so.minimize(
            self.culoss, io, method='SLSQP',  # trust-constr
            bounds=self.bounds,
            options={'disp': disp, 'maxiter': maxiter},
            #            jac=gradient_respecting_bounds(self.bounds, self.culoss),
            constraints=[
                {'type': 'eq',
                 'fun': lambda iqd: self.torque_iqd(*iqd) - torque},
                {'type': 'eq',
                 'fun': lambda iqd: np.linalg.norm(
                     self.uqd(w1, *iqd)) - u1max*np.sqrt(2)}])
        if res['success']:
            return (*res.x, self.torque_iqd(*res.x))
        logger.warning("%s: w1=%f torque=%f, u1max=%f, io=%s",
                       res['message'], w1, torque, u1max, io)
        raise ValueError(res['message'])

    def w1_umax(self, u, iq, id, iex):
        """return frequency w1 at given voltage u and id, iq current

        Keyword arguments:
        u -- the maximum voltage (RMS)
        iq, id -- the d-q currents"""
        w10 = np.sqrt(2)*u/np.linalg.norm(self.psi(iq, id, iex))
        return so.fsolve(
            lambda w1: np.linalg.norm(self.uqd(w1, iq, id, iex))-u*np.sqrt(2),
            w10)[0]

    def characteristics(self, T, n, u1max, nsamples=50):
        """calculate torque speed characteristics.
        return dict with list values of
        n, T, u1, i1, beta, cosphi, pmech, n_type

        Keyword arguments:
        T -- (float) the maximum torque in Nm
        n -- (float) the maximum speed in 1/s
        u1max -- (float) the maximum voltage in V rms
        nsamples -- (optional) number of speed samples
        """
        iq, id, iex = self.iqd_torque(T)
        w1type = self.w1_umax(u1max, iq, id, iex)
        wmType = w1type/self.p
        pmax = T*wmType

        def tload(wm):
            if abs(wm*T) < abs(pmax):
                return T
            return pmax/wm

        wmtab = []
        dw = 0
        wmMax = 3.5*wmType
        if n > 0:
            wmMmax = min(wmMax, 2*np.pi*n)
        if wmType > wmMax:
            wmrange = sorted([0, wmMax])
            wmtab = np.linspace(0, wmMax, nsamples).tolist()
        else:
            wmrange = sorted([0, wmType, wmMax])
            nx = int(round(nsamples*wmType/wmMax))
            dw = (wmMax-wmType)/(nsamples-nx)
            wmtab = (np.linspace(0, wmType, nx).tolist() +
                     np.linspace(wmType+dw, wmMax, nsamples-nx).tolist())

        logger.info("Speed range %s", wmrange)
        wmtab[0] = 0

        r = dict(u1=[], i1=[], id=[], iq=[], T=[], cosphi=[], n=[],
                 beta=[], plfe1=[], plcu1=[], plcu2=[])
        T = [tload(wx) for wx in wmtab]
        tfric = self.kfric_b*self.rotor_mass*30e-3/np.pi
        w1tab = []
        for wm, tq in zip(wmtab, T):
            try:
                w1 = wm*self.p
                if w1 <= w1type:
                    iq, id, iex = self.iqd_torque(tq)
                else:
                    iq, id, iex = self.iqd_torque_umax(
                        tq, w1, u1max,
                        (0.9*iq, 0.9*id,
                         min(self.bounds[-1][0], 0.9*iex)))[:-1]
                uq, ud = self.uqd(w1, iq, id, iex)
                u1 = np.linalg.norm((uq, ud))/np.sqrt(2)
                f1 = w1/2/np.pi
                r['id'].append(id)
                r['iq'].append(iq)
                r['u1'].append(u1)
                beta = np.arctan2(id, iq)
                if beta > 0:
                    beta -= 2*np.pi
                i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
                r['i1'].append(i1)
                r['beta'].append(beta/180*np.pi)
                gamma = np.arctan2(ud, uq)
                r['cosphi'].append(np.cos(gamma - beta))
                r['plfe1'].append(self.plfe1(beta, i1, iex, f1))
                r['plcu1'].append(self.m*i1**2*self.rstat(w1))
                r['plcu2'].append(iex**2*self.rrot(0))
                r['T'].append(tq-tfric)
                r['n'].append(wm/2/np.pi)
            except ValueError as ex:
                logger.debug("wm %f T %f", wm, tq)
                break

        r['plfe'] = r['plfe1']
        r['plcu'] = (np.array(r['plcu1']) + np.array(r['plcu2'])).tolist()
        r['plfric'] = [2*np.pi*n*tfric for n in r['n']]
        r['pmech'] = [2*np.pi*n*tq for n, tq in zip(r['n'], r['T'])]
        pmech = np.array(r['pmech'])
        pltotal = (np.array(r['plfe1']) + np.array(r['plfric']) +
                   np.array(r['plcu1']) + np.array(r['plcu2']))
        r['losses'] = pltotal.tolist()

        if pmech.any():
            p1 = pmech + pltotal
            if np.abs(pmech[0]) < 1e-12:
                r['eta'] = [0]
                i = 1
            else:
                r['eta'] = []
                i = 0
            if np.all(abs(p1[i:]) > abs(pmech[i:])):
                r['eta'] += (pmech[i:]/(p1[i:])).tolist()
            else:
                r['eta'] += (p1[i:]/pmech[i:]).tolist()

        return r


if __name__ == '__main__':
    import sys
    import json
    import femagtools.plot
    import matplotlib.pyplot as plt
    with open(sys.argv[1]) as fp:
        eecpar = json.load(fp)
    m = SynchronousMachine(eecpar)
    T = 240
    u1max = 163
    nmax = 1000
    r = m.characteristics(T, 0, u1max)
    femagtools.plot.characteristics(r, 'SynchronousMachine')
    plt.show()
